import warnings
import sys
import time
import csv

import numpy as np
from sklearn import svm
from sklearn import cross_validation
from sklearn import preprocessing as pr
from sklearn import metrics
from sklearn.ensemble import RandomForestClassifier
import nltk # for pos tags 

import preprocessing
import ngramGenerator
import features
import polarity # for affinity

def buildBagOfWords(positiveTweetsFile, neutralTweetsFile, negativeTweetsFile):
    positive = ngramGenerator.mostFreqList(positiveTweetsFile, 3000)
    negative = ngramGenerator.mostFreqList(neutralTweetsFile, 3000)
    neutral = ngramGenerator.mostFreqList(negativeTweetsFile, 3000)
        
    for w in positive:
        if w in negative+neutral : 
            positive.remove(w)
    
    for w in negative:
        if w in positive+neutral : 
            negative.remove(w)
    
    for w in neutral:
        if w in negative+positive : 
            neutral.remove(w)
    
    # equalize uni-grams sizes 
    m = min([len(positive), len(negative), len(neutral)])       
    
    positive = positive[0:m-1]
    negative = negative[0:m-1]
    neutral = neutral[0:m-1]
    
    return positive, negative, neutral

def mapTweet(tweet, positive, negative, neutral, stopWords, slangs, affinity, emoticons):
    out = []

    # since after reading in binary data and converting it into a string the line looks like:
    # b'"4","3","Mon May 11 03:17:40 UTC 2009","kindle2","tpryan","@stellargirl I loooooooovvvvvveee my Kindle2. Not that the DX is cool, but the 2 is fantastic in its own right."\n'
    # therefore start at pos 3 and end 3 chars before the end 
    tweet = tweet[2:-3]

    line = preprocessing.processTweet(tweet, stopWords, slangs)
    
    p = polarity.affinityPolarity(line, affinity)
    
    out.append(p)
    
    out.append(float(features.emoticonScore(line, emoticons))) # emo aggregate score be careful to modify weights
    out.append(float(len(features.hashtagWords(line)) / 40)) # number of hashtagged words
    out.append(float(len(line) / 140)) # for the length
    out.append(float(features.upperCase(line))) # uppercase existence : 0 or 1
    out.append(float(features.exclamationTest(line)))
    out.append(float(line.count("!") / 140))
    out.append(float((features.questionTest(line))))
    out.append(float(line.count('?') / 140))
    out.append(float(features.freqCapital(line)))
    
    u = features.scoreUnigram(line, positive, negative, neutral)
    
    out.extend(u)
    
    return out

def loadMatrix(posfilename, neufilename, negfilename, poslabel, neulabel, neglabel, positive, negative, neutral, stopWords, slangs, affinity, emoticons):
    vectors = []
    labels = []
    
    print("  Processing positive training data...")
    f = open(posfilename, 'rb')
    kpos = 0
    kneg = 0
    kneu = 0
    line = f.readline()
    while line:
        line = str(line)        
        try:
            kpos += 1
            z = mapTweet(line, positive, negative, neutral, stopWords, slangs, affinity, emoticons)
            vectors.append(z)
            labels.append(float(poslabel))
        except:
            None
        line = f.readline()
    f.close()
    
    print("  Processing neutral training data...")
    f = open(neufilename, 'rb')
    line = f.readline()
    while line:
        line = str(line)
        try:
            kneu += 1
            z = mapTweet(line, positive, negative, neutral, stopWords, slangs, affinity, emoticons)
            vectors.append(z)
            labels.append(float(neulabel))
        except:
            None
        line = f.readline()
    f.close()
    
    print("  Processing negative training data...")
    f = open(negfilename, 'rb')
    line = f.readline()
    while line:
        line = str(line)
        try:
            kneg += 1
            z = mapTweet(line, positive, negative, neutral, stopWords, slangs, affinity, emoticons)
            vectors.append(z)
            labels.append(float(neglabel))
        except:
            None
        line = f.readline()
    f.close()

    return vectors, labels

def createSvmModel(knel, c):
    clf = svm.SVC(kernel = knel, C = c)
    return clf

def createForestModel(estimators = 100):
    rf = RandomForestClassifier(n_estimators = estimators)
    return rf

# map tweet into a vector 
def trainModel(X, Y, knel, c, estimators): # relaxation parameter
    #clf = createSvmModel(knel, c)
    clf = createForestModel(estimators)
    clf.fit(X, Y)
    return clf

def predict(tweet, model, scaler, normalizer, positive, negative, neutral, stopWords, slangs, affinity, emoticons): # test a tweet against a built model 
    z = mapTweet(tweet, positive, negative, neutral, stopWords, slangs, affinity, emoticons) # mapping
    z_scaled = scaler.transform(z)
    z = normalizer.transform([z_scaled])
    z = z[0].tolist()
    return model.predict([z]).tolist()[0] # transform nympy array to list 

def loadTest(filename, scaler, normalizer, positive, negative, neutral, stopWords, slangs, affinity, emoticons): # function to load test file in the csv format : sentiment,tweet 
    f = open(filename,'rb')
    line = f.readline()
    labels = []
    vectors = []
    while line:
        line = str(line)
        # since after reading in binary data and converting it into a string the line looks like:
        # b'"4","3","Mon May 11 03:17:40 UTC 2009","kindle2","tpryan","@stellargirl I loooooooovvvvvveee my Kindle2. Not that the DX is cool, but the 2 is fantastic in its own right."\n'
        # therefore start at pos 3 and end 3 chars before the end 
        #l = line[:-1].split(r'","')
        l = line[2:-3].split(r'","')
        s = float(l[0][1:])
        tweet = l[5][:-1]

        z = mapTweet(tweet, positive, negative, neutral, stopWords, slangs, affinity, emoticons)
        z_scaled = scaler.transform(z)
        z = normalizer.transform([z_scaled])
        z = z[0].tolist()
        vectors.append(z)
        labels.append(s)
        line = f.readline()

    f.close()
    return vectors, labels

def writeTest(filename, model, scaler, normalizer, positive, negative, neutral, stopWords, slangs, affinity, emoticons): # function to load test file in the csv format : sentiment,tweet 
    f = open(filename, 'rb')
    line = f.readline()
    fo = open(filename + ".svm_result", "w")
    fo.write("old,tweet,new\n")
    while line:
        line = str(line)
        # since after reading in binary data and converting it into a string the line looks like:
        # b'"4","3","Mon May 11 03:17:40 UTC 2009","kindle2","tpryan","@stellargirl I loooooooovvvvvveee my Kindle2. Not that the DX is cool, but the 2 is fantastic in its own right."\n'
        # therefore start at pos 3 and end 3 chars before the end 
        #l = line[:-1].split(r'","')
        l = line[2:-3].split(r'","')        
        s = float(l[0][1:])
        tweet = l[5][:-1]
        nl = predict(tweet, model, scaler, normalizer, positive, negative, neutral, stopWords, slangs, affinity, emoticons)
        fo.write(r'"' + str(s) + r'","' + tweet + r'","' + str(nl) + r'"' + "\n")
        line = f.readline()
    f.close()
    fo.close()
    pass

def loadUnclassifiedTweetsCsv(filename, scaler, normalizer, positive, negative, neutral, stopWords, slangs, affinity, emoticons): # function to load test file in the csv format : sentiment,tweet 
    #f = open(filename, 'r', encoding="utf-16")
    #line = f.readline()
    #vectors = []
    #while line:

    vectors = []
    with open(filename, newline='', encoding="utf-16") as csvfile:
        reader = csv.reader(csvfile, delimiter=',', quotechar='"')
        for line in reader:
            
        #line = str(line)

            tweet = line[3]
            tweet = preprocessing.processTweet(tweet, stopWords, slangs)
    
            z = mapTweet(tweet, positive, negative, neutral, stopWords, slangs, affinity, emoticons)
            z_scaled = scaler.transform(z)
            z = normalizer.transform([z_scaled])
            z = z[0].tolist()
    
            vectors.append(z)
    
            #line = f.readline()

    csvfile.close()
    return vectors

def writeUnclassifiedTweetsCsv(filename, model, scaler, normalizer, positive, negative, neutral, stopWords, slangs, affinity, emoticons): # function to load test file in the csv format : sentiment,tweet

    tweets = []
    labels = []

    fo = open(filename + ".svm_result", "w", encoding="utf-16")
    fo.write("TWEET, CLASSIFICATION\n")
    fo.write("<< SEE ADDITIONAL INFORMATION (SENTIMENT RATIO, ...) AT THE END OF THE FILE >>\n")
    fo.write("---------------------------------------------------------------------------------\n")

    with open(filename, newline='', encoding="utf-16") as csvfile:
        reader = csv.reader(csvfile, delimiter=',', quotechar='"')
        for line in reader:
            tweet = line[3]
            tweet = preprocessing.processTweet(tweet, stopWords, slangs)
            
            nl = predict(tweet, model, scaler, normalizer, positive, negative, neutral, stopWords, slangs, affinity, emoticons)
            fo.write(r'"' + tweet + r'","' + str(nl) + r'"' + "\n")
            
            tweets.append(tweet)
            labels.append(nl)

    csvfile.close()
    
    pos, neu, neg = getTweetsRatio(labels)
    fo.write("---------------------------------------------------------------------------------\n")
    fo.write("------------------ ADDITIONAL INFORMATION ---------------------------------------\n")
    rowText = '%.2f%% POSITIVE, %.2f%% NEUTRAL, %.2f%% NEGATIVE\n' % (pos * 100.0, neu * 100.0, neg * 100.0)
    fo.write(rowText)
    fo.write("MODEL TYPE: %s\n" % (type(model)))
    localtime = time.localtime()
    timeString = time.strftime("%Y-%m-%d %H:%M:%S", localtime)
    fo.write("DATE TIME: %s\n" % (timeString))
    fo.write("---------------------------------------------------------------------------------\n")
    
    fo.close()
    
    return tweets, labels 

def getPerformance(vectors, labels, model): # for a given set of labelled vectors calculate model labels and give accuract
    a = 0 # wrong classified vectors
    newLabels = model.predict(vectors).tolist()
    acc = metrics.accuracy_score(labels, newLabels)
    pre = metrics.precision_score(labels, newLabels, pos_label = 4)
    
    #correct = 0
    #for i in range(len(newLabels)):
    #    if newLabels[i] == labels[i]:
    #        correct += 1
    #
    #perf = correct / len(newLabels) # perf == acc
    
    return acc, pre

def getTweetsRatio(labels):
    pos = 0.0
    neu = 0.0
    neg = 0.0
    
    for i in range(len(labels)):
        if labels[i] == 0.0:
            neg += 1.0
        elif labels[i] == 2.0:
            neu += 1.0
        elif labels[i] == 4.0:
            pos += 1.0
    
    pos /= len(labels)
    neu /= len(labels)
    neg /= len(labels)
    
    return pos, neu, neg

def normalizeAndScale(X):
    # features standardization 
    X_scaled=pr.scale(np.array(X))
    scaler = pr.StandardScaler().fit(X) # to use later for testing data scaler.transform(X) 
    
    # features Normalization
    X_normalized = pr.normalize(X_scaled, norm='l2') # l2 norm
    normalizer = pr.Normalizer().fit(X_scaled)  # as before normalizer.transform([[-1.,  1., 0.]]) for test
    
    X=X_normalized
    X=X.tolist()
    
    return X, scaler, normalizer

def kFoldCrossValidationSvm(X, Y, kernelFunctions, cParams, cvFolds=5):
    # 5 fold cross validation
    x=np.array(X)
    y=np.array(Y)
        
    acc = 0.0
    prec = 0.0
    cParam = 0.0
    kernelFunc = ''
    iter = 0
    
    #linear: \langle x, x'\rangle.
    #polynomial: (\gamma \langle x, x'\rangle + r)^d. d is specified by keyword degree, r by coef0.
    #rbf: \exp(-\gamma |x-x'|^2). \gamma is specified by keyword gamma, must be greater than 0.
    #sigmoid (\tanh(\gamma \langle x,x'\rangle + r)), where r is specified by coef0.
    
    coef0 = 0.0
    degree = 3
    gamma = 'auto'
    #max_iter = -1
    max_iter = 1000
    tol = 0.001
    
    for knel in kernelFunctions:
        for c in cParams:
    
            clf = createSvmModel(knel, c)
            scores = cross_validation.cross_val_score(clf, x, y, cv = cvFolds, scoring = 'accuracy')
            precisions = cross_validation.cross_val_score(clf, x, y, cv = cvFolds, scoring = 'precision')
            if (scores.mean() > acc and precisions.mean() > prec):
                acc = scores.mean()
                prec = precisions.mean()
                kernelFunc = knel
                cParam = c
            iter += 1
            print("  Iteration " + str(iter) + " : c Parameter : " + str(c) + ", Kernel : " + str(knel))
            print("    => Accuracy of the model using %i fold cross validation : %0.2f (+/- %0.2f)" % (cvFolds, scores.mean(), scores.std() * 2))# Actual testing 
            print("    => Precision of the model using %i fold cross validation : %0.2f (+/- %0.2f)" % (cvFolds, precisions.mean(), precisions.std() * 2))# Actual testing 

    return acc, prec, kernelFunc, cParam

def kFoldCrossValidationForest(X, Y, estimators, cvFolds=5):
    # 5 fold cross validation
    x=np.array(X)
    y=np.array(Y)
        
    acc = 0.0
    prec = 0.0
    estimator = 0
    iter = 0

    for est in estimators:
    
        clf = createForestModel(est)
        scores = cross_validation.cross_val_score(clf, x, y, cv = cvFolds, scoring = 'accuracy')
        precisions = cross_validation.cross_val_score(clf, x, y, cv = cvFolds, scoring = 'precision')
        if (scores.mean() > acc and precisions.mean() > prec):
            acc = scores.mean()
            prec = precisions.mean()
            estimator = est
        iter += 1
        print("  Iteration " + str(iter) + " : Estimation : " + str(est))
        print("    => Accuracy of the model using %i fold cross validation : %0.2f (+/- %0.2f)" % (cvFolds, scores.mean(), scores.std() * 2))# Actual testing 
        print("    => Precision of the model using %i fold cross validation : %0.2f (+/- %0.2f)" % (cvFolds, precisions.mean(), precisions.std() * 2))# Actual testing 

    return acc, prec, estimator

def main():    
    startTime = time.time()
    
    slangFile = 'resources/internetSlangs.txt'
    stopWordsFile = 'resources/stopWords.txt'
    affinityFile = 'resources/affinity.txt'
    emoticonsFile = 'resources/emoticons.txt'
    positiveTweetsFile = 'resources/positive1.csv'
    neutralTweetsFile = 'resources/neutral1.csv'
    negativeTweetsFile = 'resources/negative1.csv'
    
    unclassifiedTweetsFile = 'storedTweets/20151124_180107-tweets-microsoft-1000.csv'
    
    print("Start classification of \"%s\"" % (unclassifiedTweetsFile))
    
    processStepStartTime = time.time()
    print("Initializing dictionaries (stop words, slang) ...")    
    stopWords = preprocessing.getStopWordList(stopWordsFile)
    slangs = preprocessing.loadSlangs(slangFile)
    affinity = polarity.loadAffinity(affinityFile)
    emoticons = features.createEmoticonDictionary(emoticonsFile)
    print('  => Done in [s]: %.4f' % (time.time() - processStepStartTime))
    
    processStepStartTime = time.time()
    print("Creating n-grams / building bags of words...")
    positive, negative, neutral = buildBagOfWords(positiveTweetsFile, neutralTweetsFile, negativeTweetsFile)
    print('  => Done in [s]: %.4f' % (time.time() - processStepStartTime))
    
    processStepStartTime = time.time()
    print("Loading training data...")
    X, Y = loadMatrix(positiveTweetsFile, neutralTweetsFile, negativeTweetsFile, '4', '2', '0', positive, negative, neutral, stopWords, slangs, affinity, emoticons)
    print('  => Done in [s]: %.4f' % (time.time() - processStepStartTime))
    
    processStepStartTime = time.time()
    print("Scaling and normalizing data...")
    X, scaler, normalizer = normalizeAndScale(X)
    print('  => Done in [s]: %.4f' % (time.time() - processStepStartTime))    
    
    processStepStartTime = time.time()
    folds = 5
    print("Performing %i-fold cross-validation..." % (folds))
    #kernelFunctions = ['linear', 'rbf', 'poly', 'sigmoid']
    #kernelFunctions = ['linear']
    #kernelFunctions = ['linear', 'rbf', 'poly']
    #cParams = [0.01 * i for i in range(1,2)]
    #cParams = [0.01, 0.015, 0.02]
    #cParams = [0.01, 0.02, 0.1, 0.2, 1, 10, 100, 1000]
    #cParams = [0.01]
    #acc, prec, kernelFunc, cParam = kFoldCrossValidationSvm(X, Y, kernelFunctions, cParams, folds)
    #estimators = [10, 100, 250, 500, 1000]
    estimators = [100]
    acc, prec, estimator = kFoldCrossValidationForest(X, Y, estimators, folds)
    print('  => Done in [s]: %.4f' % (time.time() - processStepStartTime))
    
    processStepStartTime = time.time()
    print("Training model...")
    #kernelFunc = 'linear'
    #cParam = 0.01
    model = trainModel(X, Y, knel = None, c = None, estimators = estimator)
    print('  => Done in [s]: %.4f' % (time.time() - processStepStartTime))
    
    processStepStartTime = time.time()
    print("Loading and classifying tweets...")
    # think about removing the search term before processing the tweets.
    # i.e. if the term is "star wars" then wars could lead to negative classification.
    
    #vectors = loadUnclassifiedTweetsCsv(unclassifiedTweetsFile, termToRemove, scaler, normalizer, positive, negative, neutral, stopWords, slangs, affinity, emoticons)
    tweets, labels = writeUnclassifiedTweetsCsv(unclassifiedTweetsFile, model, scaler, normalizer, positive, negative, neutral, stopWords, slangs, affinity, emoticons)
    pos, neu, neg = getTweetsRatio(labels)
    print('  Ratio of positive, neutral and negative tweets: %.2f%% (pos), %.2f%% (neu), %.2f%% (neg)' % (pos * 100.0, neu * 100.0, neg * 100.0))
    print('  => Done in [s]: %.4f' % (time.time() - processStepStartTime))
    
    print('\nDONE.')
    print('Total time [s]: %.4f' % (time.time() - startTime))
    
    pass

if __name__ == '__main__':
    warnings.simplefilter('ignore') # ignore all warnings
    #warnings.filterwarnings('ignore', category=UserWarning, append=False)
    #warnings.filterwarnings('ignore', category=DeprecationWarning, append=False)
    #warnings.filterwarnings('ignore', category=FutureWarning, append=False)
    
    sys.exit(main())
    